import cv2
import numpy


def select_image_range(image):
    # 定义变量和鼠标回调函数
    ix, iy = -1, -1
    on_drawing = False
    roi = None

    # 创建临时图像并复制原图
    temp_image = image.copy()
    winname = 'image'

    def mouse_callback(event, x, y, flags, param):
        nonlocal ix, iy, on_drawing, roi, temp_image
        # 如果鼠标左键按下，则开始绘制ROI框架
        if event == cv2.EVENT_LBUTTONDOWN:
            on_drawing = True
            ix, iy = x, y
        # 绘制ROI框架
        elif on_drawing and event == cv2.EVENT_MOUSEMOVE:
            mask = cv2.rectangle(numpy.zeros_like(temp_image), (ix, iy), (x, y), (0, 255, 0), 1)
            cv2.imshow(winname, temp_image | mask)
        # 鼠标左键释放，绘制ROI框架并提取ROI
        elif event == cv2.EVENT_LBUTTONUP:
            on_drawing = False
            # cv2.rectangle(temp_image, (ix, iy), (x, y), (0, 255, 0), 2)
            roi = ix, iy, x, y
            print(roi)

    # 创建窗口并绑定鼠标回调函数
    cv2.namedWindow(winname)
    cv2.setMouseCallback(winname, mouse_callback)
    # 显示截图

    cv2.imshow(winname, temp_image)
    if cv2.waitKey(0) == 13:  # 按下回车键结束循环
        # 销毁所有窗口
        cv2.destroyAllWindows()
        return roi



import ast

import astunparse


class FileFunction:
    def __init__(self, pyfile: str):
        """
        Convert a python file into an executable function,
        and the passed parameters will modify the same name variables during execution. \n
        1, Only keyword parameter passing is supported.  \n
        2, To define the RETURN variable in the original python file to return a value. \n
        3, The passed parameters is const and can not be modified. \n
        :param pyfile: original python file.
        """
        self.pyfile = pyfile
        with open(self.pyfile) as f:
            self.code_tree = ast.parse(f.read())

    def __call__(self, **kwargs):
        """
        1, Only keyword parameter passing is supported.  \n
        2, To define the RETURN variable in the original python file to return a value. \n
        3, The passed parameters is const and can not be modified. \n
        """
        for code in self.code_tree.body:
            locals().update(kwargs)
            exec(astunparse.unparse(code), locals())

        return locals().get("RETURN", None)



import os

import cv2
import einops
import numpy


def write_image(image: numpy.ndarray, filename: str, *, resize=None, rgb_mode=True) -> None:
    """
    将图像写入文件。
    :param image: numpy.array, 输入图像。
    :param filename: str, 输出文件名。
    :param resize: (h, w)，默认为None
    :param rgb_mode: cv2是以BGR的形式写入，如果当前是RGB需要转化为RGB，默认为True。
    """
    if rgb_mode:
        image = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)
    if resize:
        image = cv2.resize(image, resize)
    cv2.imwrite(filename, image)

    print(filename, "written")


def write_video(video: numpy.ndarray, filename: str, *, fps: int = 30, resize=None, rgb_mode=True) -> None:
    """
    将cv2读取的维度格式FHWC（frame_count, height, width, channel_count）视频帧写入文件。
    如果缺少channel_count维度，则会repeat补上。
    函数接受一个视频帧、一个输出文件名和一个视频帧率（默认为30帧/秒），

    使用OpenCV的VideoWriter函数创建视频写入对象，然后逐帧将视频写入文件。
    最后，释放视频写入对象。注意，为了使写入的视频具有更好的兼容性，我们选择了MP4编码格式。

    :param video: numpy.array, 输入视频帧。
    :param filename: str, 输出文件名。
    :param fps: int, 输出视频帧率。
    :param resize: (h, w)，默认为None
    :param rgb_mode: cv2是以BGR的形式写入，如果当前是RGB需要转化为RGB，默认为True。
    """
    video = numpy.array(video, dtype=numpy.uint8)
    if video.ndim == 4:
        frame_count, height, width, channel_count = video.shape
    elif video.ndim == 3:
        video = einops.repeat(video, "f h w -> f h w c", c=3)
        frame_count, height, width, channel_count = video.shape
    else:
        raise Exception(f"video dim is not (f h w [c]).")

    _, ext = os.path.splitext(filename)
    fourcc_dict = {'.mp4': 'mp4v', '.avi': 'XVID', '.mkv': 'avc1', '.mov': 'mp4v', '.3gp': 'mp4v', '.webm': 'VP80',
                   '.wmv': 'WMV1'}
    fourcc = cv2.VideoWriter_fourcc(*fourcc_dict[ext])  # 定义输出视频编码格式

    writer = cv2.VideoWriter(filename, fourcc, fps, (width, height))  # 创建视频写入对象

    for frame in video:
        if rgb_mode:
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2BGR)
        if resize:
            frame = cv2.resize(frame, resize)
        writer.write(frame)  # 逐帧写入视频

    writer.release()  # 释放视频写入对象

    print(filename, "written")



import glob
import pickle
from builtins import ValueError
from functools import lru_cache
from os.path import join
from typing import Tuple

import einops
import numpy
from torch.utils.data import dataset


def unpickle(file):
    with open(file, 'rb') as fo:
        obj = pickle.load(fo)
    return obj


def load_databatch(data_file, img_size=64):
    # data_file = os.path.join(data_folder, f'train_data_batch_{str(idn)}')
    d = unpickle(data_file)
    x = numpy.array(d['data'], dtype="float32")
    y = numpy.array(d['labels'], dtype="int32")
    # mean = numpy.array(d['mean'])

    # Labels are indexed from 1, shift it so that indexes start at 0
    y = y - 1

    x = einops.rearrange(x, "b (c h w) -> b c h w", h=img_size, w=img_size)
    # mean = einops.rearrange(mean, "(c h w) -> h w c", h=img_size, w=img_size)

    return dict(x=x, y=y)


class ImageNet64(dataset.Dataset):
    BLOCK_CACHE_SIZE = 2
    DATA_CACHE_SIZE = 100_000

    def __init__(self, root, train=True):
        """
        ImageNet64 的数据集，用于加载 64x64 像素的图像数据集。
        数据集被分成了 10 个文件（train_data_batch_1 至 train_data_batch_10），
        其中每个文件包含了相同数量的图像数据（128,116 或 128,123 张）。
        每个批次都保存在 pickle 文件中。该数据集支持使用索引直接访问单个图像和标签，
        并且使用缓存来尽可能快地访问已加载的数据块。
        :param root: train_data_batch_* 文件的父文件夹
        """
        super().__init__()
        self.root = root

        if train:
            if len(glob.glob(join(root, "train_data_batch_*"))) != 10:
                raise FileNotFoundError(f"train_data_batch "
                                        f"file count {len(glob.glob(join(root, 'train_data_batch_*')))} "
                                        f"is not equals to 10.")

            each_block_length = [128116, 128116, 128116, 128116, 128116,
                                 128116, 128116, 128116, 128116, 128123]

            # self.index_block_map = \
            #     [(i, (sum(each_block_length[:i - 1]), sum(each_block_length[:i]))) for i in range(1, 11)]
            self.index_block_map = \
                [(1, (0, 128116)), (2, (128116, 256232)), (3, (256232, 384348)), (4, (384348, 512464)),
                 (5, (512464, 640580)), (6, (640580, 768696)), (7, (768696, 896812)), (8, (896812, 1024928)),
                 (9, (1024928, 1153044)), (10, (1153044, 1281167))]

            # print(self.index_block_map)
            self.data_length = sum(each_block_length)

            self.get_data = self.get_data_train
        else:
            if len(glob.glob(join(root, "val_data"))) != 1:
                raise FileNotFoundError(f"val_data file not exists.")
            self.data_list = load_databatch(join(root, "val_data"))
            self.data_length = len(self.data_list["y"])

            self.get_data = self.get_data_val

    @lru_cache(BLOCK_CACHE_SIZE)
    def get_data_block(self, block_idn):
        # print(block_idn)
        return load_databatch(data_file=join(self.root, f'train_data_batch_{str(block_idn)}'))

    @lru_cache(DATA_CACHE_SIZE)
    def get_data_train(self, index):
        if index >= self.data_length or index < 0:
            raise IndexError(f"index {index} out of bounds for size [0, {self.data_length})")

        count_previous_blocks_length = 0
        for map_block_idn, map_interval in self.index_block_map:
            # print(map_block_idn)
            if map_interval[0] <= index < map_interval[1]:
                block_idn = map_block_idn
                break
            else:
                count_previous_blocks_length = map_interval[1]
        else:
            raise ValueError(f"没有找到{index}")

        data_block = self.get_data_block(block_idn)
        # print(block_idn, index - count_previous_blocks_length)
        return data_block["x"][index - count_previous_blocks_length], \
            data_block["y"][index - count_previous_blocks_length]

    def get_data_val(self, index):
        return self.data_list["x"][index], self.data_list["y"][index]

    def __getitem__(self, index) -> Tuple[numpy.ndarray, int]:
        return self.get_data(index)

    def __len__(self):
        return self.data_length



from typing import List

import matplotlib.pyplot as plt
import numpy
from fastcore.basics import flatten
from matplotlib.axes import Axes
from matplotlib.figure import Figure


def show_multi_images_in_matplotlib(images: List[numpy.array], subtitles=None, col_num=3, title=None, show=False):
    col_num = min(len(images), col_num)
    images_count = len(images)

    if not subtitles:
        subtitles = [None] * images_count

    if images_count > len(subtitles):
        subtitles.extend([None] * (images_count - len(subtitles)))

    subtitles = subtitles[:images_count]

    row_num = -(-images_count // col_num)  # -a//b表示将a除以b并向下取整，最后将结果取反，即可得到向上取整的效果
    fig, axs = plt.subplots(nrows=row_num, ncols=col_num, squeeze=False)  # type: Figure, Axes
    fig.suptitle(title)
    fig.set_size_inches(4 * col_num, 3.5 * row_num)

    axs_list = list(flatten(axs))
    for ax, img, subtitle in zip(axs_list[:images_count], images, subtitles):  # type: Axes
        ax.set_title(subtitle)
        ax.imshow(img)

    for ax in axs_list[images_count:]:
        ax.axis('off')

    fig.tight_layout()

    if show:
        plt.show()

    return fig



import os
import random

import numpy
import torch
from torch.utils.data import Dataset


class ListDataset(Dataset):
    def __init__(self, data_list):
        self.data_list = data_list

    def __len__(self):
        return len(self.data_list)

    def __getitem__(self, index):
        return self.data_list[index]


def print_support_devices():
    result = dict()
    for has_x in filter(lambda x: str.startswith(x, "has_"), dir(torch)):
        method = getattr(torch, has_x)
        print(has_x, method)
        result[has_x[4:]] = method


def seed_everything(seed=42):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    numpy.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True



import numpy


def crop_image(image: numpy.ndarray,
               x1: int, y1: int, x2: int, y2: int,
               h_axis: int = 0, w_axis: int = 1) -> numpy.ndarray:
    """
    可对图像(h w [c])进行截取，返回截取后的数组。
    :param image: numpy.array, 输入图像或视频的帧。
    :param x1: int, 左上角 x 坐标。
    :param y1: int, 左上角 y 坐标。
    :param x2: int, 右下角 x 坐标。
    :param y2: int, 右下角 y 坐标。
    :param h_axis: int, 数组中表示高度（height）的轴。
    :param w_axis: int, 数组中表示宽度（width）的轴。

    :return cropped_image: numpy.array, 截取后的二维或三维数组。
    """
    # 检查 h_axis 和 w_axis 是否超出数组维度
    if h_axis >= image.ndim or w_axis >= image.ndim:
        raise ValueError(f"h_axis:{h_axis} >= ndim:{image.ndim} or w_axis:{w_axis} >= ndim:{image.ndim}")

    # 修正 x1, y1, x2, y2，使得坐标表示从左到右，从上到下
    x1, x2 = sorted([x1, x2])
    y1, y2 = sorted([y1, y2])

    # 创建用于截取数组的切片
    slices = [slice(None, None, None) for _ in range(image.ndim)]
    slices[h_axis] = slice(y1, y2)
    slices[w_axis] = slice(x1, x2)

    # 截取图片
    cropped_image = image[tuple(slices)]

    return cropped_image


def crop_video(video: numpy.ndarray, x1: int, y1: int, x2: int, y2: int, h_axis: int = 1,
               w_axis: int = 2) -> numpy.ndarray:
    """
    对视频(f h w [c])中的每一帧进行截取，返回截取后的视频。
    实际上是将视频看成一个多维图像，直接采用`crop_image`函数对多维图像进行切割。

    :param video: numpy.array, 输入视频。
    :param x1: int, 左上角 x 坐标。
    :param y1: int, 左上角 y 坐标。
    :param x2: int, 右下角 x 坐标。
    :param y2: int, 右下角 y 坐标。
    :param h_axis: int, 数组中表示高度（height）的轴。
    :param w_axis: int, 数组中表示宽度（width）的轴。
    :return cropped_video: numpy.array, 截取后的视频。
    """

    return crop_image(video, x1, y1, x2, y2, h_axis, w_axis)



from typing import Optional, List

import cv2
import imagesize
import numpy


def get_image_size(image_path):
    """
    :param image_path: 图像路径
    :return: height, width
    """

    width, height = imagesize.get(image_path)

    if (height, width) == (-1, -1):
        image: numpy.ndarray = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)
        # 获取图像尺寸
        height, width = image.shape

    return height, width


def read_image(image_path, grey: bool = False, resize=None, rgb_mode: bool = True):
    """
    读取图像函数

    :param image_path: 图像路径
    :param grey: 是否将图像转换为灰度图像，默认为False
    :param resize: (h, w)，默认为None
    :param rgb_mode: 使用opencv读出的图像是BGR格式，是否将图像转换为RGB格式，默认为True
    :return: 读取到的图像数组
    """

    if grey:
        # 读取灰度图像
        image = cv2.imread(image_path, cv2.IMREAD_GRAYSCALE)
    else:
        # 读取彩色图像
        image = cv2.imread(image_path)
        if rgb_mode:
            # 将BGR格式转换为RGB格式
            image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

    if resize:
        image = cv2.resize(image, resize)

    return image


def read_images(image_paths: List[str], grey: bool = False, resize=None, rgb_mode: bool = True):
    image_list = []
    for image_path in image_paths:
        image = read_image(image_path, grey, resize, rgb_mode)
        image_list.append(image)

    return numpy.stack(image_list, axis=0)


def get_video_info(video_path):
    """
    获取帧速率、帧数、宽度和高度
    :param video_path: 视频文件地址
    :return: 帧数、宽度、高度和帧速率。如果视频文件无法打开，则返回None。
    """

    # 打开视频文件
    cap = cv2.VideoCapture(video_path)

    # 检查视频文件是否成功打开
    if not cap.isOpened():
        return None

    # 获取帧速率、帧数、宽度和高度
    frames_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))
    fps = int(cap.get(cv2.CAP_PROP_FPS))
    height = int(cap.get(cv2.CAP_PROP_FRAME_HEIGHT))
    width = int(cap.get(cv2.CAP_PROP_FRAME_WIDTH))

    # 释放VideoCapture对象
    cap.release()

    # 返回帧数、宽度、高度和帧速率
    return frames_count, fps, height, width


def read_video(video_path, *,
               specific_frame: Optional[List[int]] = None,
               start: int = 0, stop: int = -1, step: int = 1,
               resize=None,
               rgb_mode=True):
    """
    这个函数可以读取视频文件，并返回指定范围内的帧。

    :param video_path: 视频文件路径
    :param specific_frame: 指定特定帧
    :param start: 要读取的起始帧索引，负数代表倒数（默认为 0）
    :param stop: 要读取的结束帧索引，负数代表倒数（默认为 -1，即读取到最后一帧）
    :param step: 读取帧的步长（默认为 1）
    :param resize: (h, w)，默认为None
    :param rgb_mode: 是否转化为RGB图像（默认为 True）
    :return: 读取成功返回一个代表对应帧所组成的numpy数组(h w c)，读取失败返回None
    """
    if specific_frame:
        print("Because the specific_frame has been specified, parameters start, stop and step are ignored.")

    cap = cv2.VideoCapture(video_path)

    if not cap.isOpened():
        return None

    # 结束帧数不应超过总帧数
    frames_count = int(cap.get(cv2.CAP_PROP_FRAME_COUNT))

    video = []
    if specific_frame:
        specific_frame = sorted(specific_frame)
        specific_frame.insert(0, 0)
        # for i in range(specific_frame[0]):
        #     print("ignore", i)
        #     cap.grab()

        for j in range(1, len(specific_frame)):
            for i in range(specific_frame[j] - specific_frame[j - 1]):
                ret = cap.grab()
            else:
                ret, frame = cap.read()
                if ret:
                    if rgb_mode:
                        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    if resize:
                        frame = cv2.resize(frame, resize)
                    video.append(frame)

            if not ret:
                break

    else:
        start = start % frames_count
        stop = stop % frames_count

        reverse_flag = False
        if start > stop:
            start, stop = sorted([start, stop])
            reverse_flag = not reverse_flag

        if step < 0:
            step = abs(step)
            reverse_flag = not reverse_flag

        print(start, stop, step, reverse_flag)
        # 按给定的范围读取视频帧
        for i in range(start):
            cap.grab()

        for i in range(stop - start + 1):
            # 只选取其中符合特定步长的帧
            if i % step == 0:
                ret, frame = cap.read()
                if ret:
                    if rgb_mode:
                        frame = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
                    if resize:
                        frame = cv2.resize(frame, resize)
                    video.append(frame)
            else:
                ret = cap.grab()

            if not ret:
                break

        if reverse_flag:
            video = video[::-1]

    cap.release()

    return numpy.stack(video, axis=0)



from setuptools import setup, find_packages


def read(filename):
    with open(filename, encoding='utf-8') as f:
        return f.read()


setup(name='tusuan',  # 包名
      python_requires='>=3.8.0',  # python环境
      version='0.0.9.2',  # 包的版本
      description="useful functions.",  # 包简介，显示在PyPI上

      long_description=read('README.md'),  # 读取的Readme文档内容，一整块字符串
      long_description_content_type="text/markdown",  # 指定包文档格式为markdown

      author="tusuan",  # 作者相关信息
      author_email='btk@qq.com',
      url='https://github.com/tusuan',

      # 指定包信息，还可以用find_packages()函数  # find_packages(where="./", include=["tusuan*"]),
      # packages=["tusuan",
      #           "tusuan.image_video_utils"],
      packages=find_packages(where=".", include=["tusuan*"]),
      install_requires=read('requirements.txt').splitlines(),  # 指定需要安装的依赖, 需要是一个列表

      include_package_data=True,  # 不知道做啥的
      license="MIT",
      keywords=['tusuan'],
      classifiers=[  # 一些网站的分类信息，方便用户检索
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Natural Language :: English',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10'
      ])

"""
rm -rf ./dist
pipreqs ./ --encoding=utf8 --force --mode no-pin
python3 setup.py sdist 
twine upload --repository tusuan@testpypi dist/*
pip install -i https://test.pypi.org/simple/ -U tusuan

twine upload --repository tusuan@pypi dist/*
"""
